textview.GtkSourceMap {
  font-family: BuilderBlocks;
  font-size: 1.75pt;
  line-height: 3px;
  color: alpha(currentColor, 0.75);
}
textview.GtkSourceMap:dir(ltr) {
  border-left: 1px solid @borders;
}
textview.GtkSourceMap:dir(rtl) {
  border-right: 1px solid @borders;
}

GtkSourceHoverDisplay {
  padding: 6px;
}
GtkSourceAssistant.hover-assistant {
  background: @popover_bg_color;
  color: @popover_fg_color;
  box-shadow: 0 3px 14px 7px @shadow_color,
              0 0 0 1px mix(@popover_bg_color, @popover_fg_color, .15);
}
