/* gbp-top-match-completion-filter.h
 *
 * Copyright 2022 Christian Hergert <chergert@redhat.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <gtksourceview/gtksource.h>

G_BEGIN_DECLS

#define GBP_TYPE_TOP_MATCH_COMPLETION_FILTER (gbp_top_match_completion_filter_get_type())

G_DECLARE_FINAL_TYPE (GbpTopMatchCompletionFilter, gbp_top_match_completion_filter, GBP, TOP_MATCH_COMPLETION_FILTER, GObject)

GbpTopMatchCompletionFilter *gbp_top_match_completion_filter_new            (GtkSourceCompletionProvider *provider,
                                                                             GListModel                  *model);
GtkSourceCompletionProvider *gbp_top_match_completion_filter_get_provider   (GbpTopMatchCompletionFilter *self);
void                         gbp_top_match_completion_filter_set_model      (GbpTopMatchCompletionFilter *self,
                                                                             GListModel                  *model);
void                         gbp_top_match_completion_filter_set_typed_text (GbpTopMatchCompletionFilter *self,
                                                                             const char                  *word);

G_END_DECLS
